from .widgets import *

from .. import Timecourse

def tc_record_activity(*, start=1, stop=10, gene_names=None, verbose=False):
    tc = Timecourse(start=start, stop=stop)

    state_record = []     # (tp_name, state)

    time_points = {}      # time_point_name => index
    all_tissues = set()   # all tissues across all time points

    tc.run()

    for n, state in enumerate(iter(tc)):
        tp = f"t={state.time}"
        if verbose:
            print(tp)
        time_points[tp] = n

        for ti in state.tissues:
            all_tissues.add(ti.name)
            present = state[ti]
            if verbose:
                print(f"\ttissue={ti.name}, {present.report_activity()}")

        state_record.append((tp, state))

    def is_active(tissue_name, time_point, gene):
        time_idx = time_points[time_point]
        ga = state_record[time_idx]
        active = ga[1].get_by_tissue_name(tissue_name).is_active(gene)
        return bool(active)
    
    return state_record, list(all_tissues), is_active



import pytest

import dinkum
from dinkum.vfg import Gene, Receptor, Ligand
from dinkum.vfn import Tissue
from dinkum import Timecourse
from dinkum import observations


def test_signaling_orig_api():
    dinkum.reset()

    #observations.check_is_present(gene='R', tissue='M', time=2)
    observations.check_is_present(gene='X', tissue='N', time=2)
    observations.check_is_not_present(gene='Y', tissue='M', time=2)
    observations.check_is_present(gene='Y', tissue='M', time=4)

    m = Tissue(name='M')
    n = Tissue(name='N')

    m.add_neighbor(neighbor=n)
    assert n in m.neighbors     # should this be bidirectional? probably.

    x = Gene(name='X')
    a = Gene(name='A')
    r = Receptor(name='R')
    y = Gene(name='Y')

    r.ligand(activator=a, ligand=x)

    y.activated_by(source=r)

    # receptor is always present in M
    m.add_gene(gene=a, start=1)

    # x is present in N at time >= 2
    n.add_gene(gene=x, start=2)

    dinkum.run(1, 5)


def test_signaling_new_api():
    # use newer API for Receptor
    dinkum.reset()

    #observations.check_is_present(gene='R', tissue='M', time=2)
    observations.check_is_present(gene='X', tissue='N', time=2)
    observations.check_is_not_present(gene='Y', tissue='M', time=2)
    observations.check_is_present(gene='Y', tissue='M', time=4)

    m = Tissue(name='M')
    n = Tissue(name='N')

    m.add_neighbor(neighbor=n)
    assert n in m.neighbors     # should this be bidirectional? probably.

    x = Gene(name='X')
    a = Gene(name='A')
    r = Receptor(name='R', ligand=x)
    y = Gene(name='Y')

    r.ligand(activator=a)

    y.activated_by(source=r)

    # receptor is always present in M
    m.add_gene(gene=a, start=1)

    # x is present in N at time >= 2
    n.add_gene(gene=x, start=2)

    dinkum.run(1, 5)


def test_signaling_new_api_3():
    dinkum.reset()

    #observations.check_is_present(gene='R', tissue='M', time=2)
    observations.check_is_present(gene='X', tissue='N', time=2)
    observations.check_is_not_present(gene='Y', tissue='M', time=2)
    observations.check_is_present(gene='Y', tissue='M', time=4)

    m = Tissue(name='M')
    n = Tissue(name='N')

    m.add_neighbor(neighbor=n)
    assert n in m.neighbors     # should this be bidirectional? probably.

    x = Ligand(name='X')
    a = Gene(name='A')
    r = Receptor(name='R', ligand=x)
    y = Gene(name='Y')

    r.ligand(activator=a)

    y.activated_by(source=r)    # @CTB update.

    # receptor is always present in M
    m.add_gene(gene=a, start=1)

    # x is present in N at time >= 2
    n.add_gene(gene=x, start=2)

    dinkum.run(1, 5)


def test_community_effect():
    # transient input in one cell => mutual lock on via positive feedback/
    # signalling
    # good question for students: how long does pulse need to be, and why??
    # CTB: _active_ vs _present_
    dinkum.reset()

    observations.check_is_present(gene='A', tissue='M', time=6)
    observations.check_is_not_present(gene='A', tissue='M', time=7)

    observations.check_is_not_present(gene='L', tissue='M', time=1) # not act
    observations.check_is_present(gene='L', tissue='M', time=2) # active
    observations.check_is_present(gene='R', tissue='N', time=3) # active

    observations.check_is_present(gene='Y', tissue='N', time=4)
    observations.check_is_present(gene='L', tissue='N', time=5)

    observations.check_is_present(gene='R', tissue='M', time=6) # active

    # both on (& staying on)
    observations.check_is_present(gene='Y', tissue='N', time=7)
    observations.check_is_present(gene='Y', tissue='M', time=7)

    # two tissues
    m = Tissue(name='M')
    n = Tissue(name='N')

    # neighbors
    m.add_neighbor(neighbor=n)
    n.add_neighbor(neighbor=m)
    assert n in m.neighbors     # should this be bidirectional? probably. CTB.
    assert m in n.neighbors

    # VFN:
    a = Gene(name='A')          # transient input in M to turn on ligand L
    m.add_gene(gene=a, start=1, duration=6)

    b = Gene(name='B')          # permanent input in M and N to turn on receptor R
    m.add_gene(gene=b, start=1)
    n.add_gene(gene=b, start=1)

    # VFG: 
    ligand = Gene(name='L')          # ligand
    r = Receptor(name='R')      # receptor
    y = Gene(name='Y')          # activated by R

    ligand.activated_or(sources=[a, y])

    r.ligand(activator=b, ligand=ligand) # expression driven by B,
                                         # activated by ligand

    y.activated_by(source=r)    # transcription of Y turned on by activated R.

    # so,
    # pulse of A in M turns on L in M.
    # receptor R is always expressed in N b/c B turns it on.
    # M is neighbor of N, so N sees L.
    # R is activated by L in N.
    # downstream of R, R activates L in N.
    # L then activates R in M.
    # R in M then activates L in M.

    dinkum.run(1, 12)



"""encode view-from-nucleus rules

X is-present in celltype/tissue/compartment M at time T
"""
from functools import total_ordering

_tissues = []
def _add_tissue(t):
    global _tissues
    _tissues.append(t)

def get_tissues():
    return list(sorted(_tissues))

def get_tissue_names():
    return list(sorted([ t.name for t in _tissues ]))

def reset():
    global _tissues
    _tissues = []


@total_ordering
class Tissue:
    def __init__(self, *, name=None):
        assert name, "Tissue must have a name"
        self.name = name
        self.present = []
        self.neighbors = set()

        _add_tissue(self)

    def __eq__(self, other):
        return self.name == other.name

    def __ne__(self, other):
        return self.name != other.name

    def __lt__(self, other):
        return self.name < other.name

    def __hash__(self):
        return hash(self.name)

    def add_gene(self, *, gene=None, start=None, duration=None):
        assert gene
        assert start is not None
        self.present.append((gene, start, duration))

    def add_neighbor(self, *, neighbor=None):
        assert neighbor
        self.neighbors.add(neighbor)

    def all_active(self, *, at=None):
        assert at is not None
        for gene, start, duration in self.present:
            if at >= start:
                if duration is None:
                    yield gene
                elif at < start + duration:
                    yield gene



import pytest

import dinkum
from dinkum.vfg import Gene
from dinkum.vfn import Tissue
from dinkum import Timecourse
from dinkum import observations


def test_custom_1():
    dinkum.reset()

    x = Gene(name='X')
    y = Gene(name='Y')
    m = Tissue(name='M')

    def activator_fn(X):
        return X

    x.is_present(where=m, start=1, duration=1)
    y.custom_activation(state_fn=activator_fn, delay=1)

    # set observations
    observations.check_is_present(gene='X', time=1, tissue='M')
    observations.check_is_not_present(gene='X', time=2, tissue='M')
    observations.check_is_not_present(gene='Y', time=1, tissue='M')
    observations.check_is_present(gene='Y', time=2, tissue='M')

    observations.check_is_not_present(gene='X', time=3, tissue='M')
    observations.check_is_not_present(gene='Y', time=3, tissue='M')

    # run time course
    tc = dinkum.run(1, 5)
    assert len(tc) == 5


def test_custom_1_fail():
    dinkum.reset()

    x = Gene(name='X')
    y = Gene(name='Y')
    m = Tissue(name='M')

    def activator_fn(Z):
        return X

    x.is_present(where=m, start=1, duration=1)
    y.custom_activation(state_fn=activator_fn, delay=1)

    # run time course
    with pytest.raises(Exception): # @CTB change to Dinkum exception
        tc = dinkum.run(1, 5)



"""encode view-from-genome rules.

X binds-and-upregulates Y
X binds-and-represses Y
X directly-or-indirectly-upregulates Y
X directly-or-indirectly-represses Y

X binds-and-upregulates Y if A else binds-and-represses
"""
from functools import total_ordering
import inspect

_rules = []
_genes = []

def _add_rule(ix):
    _rules.append(ix)

def get_rules():
    return list(_rules)

def get_gene_names():
    return [ g.name for g in sorted(_genes) ]

def reset():
    global _rules
    global _genes
    _rules = []
    _genes = []


def _retrieve_ligands(timepoint, states, tissue, delay):
    "Retrieve all ligands in neighboring tissues for the given timepoint/delay"
    #assert isinstance(tissue, Tissue)

    ligands = set()
    for gene in _genes:
        if gene._is_ligand:
            for neighbor in tissue.neighbors:
                if states.is_active(timepoint, delay, gene, neighbor):
                    ligands.add(gene)

    return ligands


class Interactions:
    def check_ligand(self, timepoint, states, tissue, delay):
        if getattr(self, 'ligand', None):
            ligands_in_neighbors = _retrieve_ligands(timepoint, states,
                                                     tissue, delay)
            if self.ligand in ligands_in_neighbors:
                return True
            return False
        else:
            return True         # by default, not ligand => is active


class Interaction_Activates(Interactions):
    def __init__(self, *, source=None, dest=None, delay=1):
        assert isinstance(source, Gene), f"'{source}' must be a Gene (but is not)"
        assert isinstance(dest, Gene), f"'{dest}' must be a Gene (but is not)"
        self.src = source
        self.dest = dest
        self.delay = delay

    def advance(self, *, timepoint=None, states=None, tissue=None):
        """
        The gene is active if its source was active 'delay' ticks ago.
        """
        assert states
        assert tissue
        assert timepoint is not None

        if states.is_active(timepoint, self.delay, self.src, tissue) and \
           self.check_ligand(timepoint,
                             states,
                             tissue,
                             self.delay):
            yield self.dest, 1
        else:
            yield self.dest, 0


class Interaction_Or(Interactions):
    def __init__(self, *, sources=None, dest=None, delay=1):
        for g in sources:
            assert isinstance(g, Gene), f"source '{g}' must be a Gene"
        assert isinstance(dest, Gene), f"dest '{dest}' must be a Gene"
        self.sources = sources
        self.dest = dest
        self.delay = delay

    def advance(self, *, timepoint=None, states=None, tissue=None):
        """
        The gene is active if any of its sources were activate 'delay'
        ticks ago.
        """
        assert states
        assert tissue

        source_active = [ states.is_active(timepoint, self.delay, g, tissue)
                          for g in self.sources ]

        if any(source_active) and self.check_ligand(timepoint,
                                                    states,
                                                    tissue,
                                                    self.delay):
            yield self.dest, 1
        else:
            yield self.dest, 0


class Interaction_AndNot(Interactions):
    def __init__(self, *, source=None, repressor=None, dest=None, delay=1):
        self.src = source
        self.repressor = repressor
        self.dest = dest
        self.delay = delay

    def advance(self, *, timepoint=None, states=None, tissue=None):
        """
        The gene is active if its activator was active 'delay' ticks ago,
        and its repressor was _not_ active then.
        """
        assert states
        assert tissue

        src_is_active = states.is_active(timepoint, self.delay,
                                         self.src, tissue)
        repressor_is_active = states.is_active(timepoint, self.delay,
                                              self.repressor, tissue)

        if src_is_active and not repressor_is_active and \
           self.check_ligand(timepoint,
                             states,
                             tissue,
                             self.delay):
            yield self.dest, 1
        else:
            yield self.dest, 0


class Interaction_And(Interactions):
    def __init__(self, *, sources=None, dest=None, delay=1):
        self.sources = sources
        self.dest = dest
        self.delay = delay

    def advance(self, *, timepoint=None, states=None, tissue=None):
        """
        The gene is active if all of its sources were active 'delay' ticks
        ago.
        """
        assert states
        assert tissue

        source_active = [ states.is_active(timepoint, self.delay, g, tissue)
                          for g in self.sources ]

        if all(source_active) and self.check_ligand(timepoint,
                                                    states,
                                                    tissue,
                                                    self.delay):
            yield self.dest, 1
        else:
            yield self.dest, 0


class Interaction_ToggleRepressed(Interactions):
    def __init__(self, *, tf=None, cofactor=None, dest=None, delay=1):
        self.tf = tf
        self.cofactor = cofactor
        self.dest = dest
        self.delay = delay

    def advance(self, *, timepoint=None, states=None, tissue=None):
        """
        The gene is active if the tf was active and the cofactor was active
        'delay' ticks ago.
        """
        assert states
        assert tissue

        tf_active = states.is_active(timepoint, self.delay,
                                     self.tf, tissue)
        cofactor_active = states.is_active(timepoint, self.delay,
                                           self.cofactor, tissue)


        if tf_active and not cofactor_active and self.check_ligand(timepoint,
                                                                   states,
                                                                   tissue,
                                                                   self.delay):
            yield self.dest, 1
        else:
            yield self.dest, 0


class Interaction_Arbitrary(Interactions):
    def __init__(self, *, dest=None, state_fn=None, delay=1):
        assert dest
        assert state_fn

        self.dest = dest
        self.state_fn = state_fn
        self.delay = delay

    def advance(self, *, timepoint=None, states=None, tissue=None):
        assert states
        assert tissue

        dep_gene_names = inspect.getfullargspec(self.state_fn).args
        dep_genes = []
        for name in dep_gene_names:
            found = False
            for g in _genes:
                if g.name == name:
                    dep_genes.append(g)
                    found = True
                    break
            if not found:
                raise Exception(f"no such gene: '{name}'")

        dep_state = [ states.is_active(timepoint, self.delay, g, tissue)
                      for g in dep_genes ]
        is_active = self.state_fn(*dep_state)

        if is_active and self.check_ligand(timepoint,
                                           states,
                                           tissue,
                                           self.delay):
            yield self.dest, 1
        else:
            yield self.dest, 0


class Interaction_Ligand(Interactions):
    def __init__(self, *, activator=None, ligand=None, receptor=None, delay=1):
        self.activator = activator
        self.ligand = ligand
        self.receptor = receptor
        self.delay = delay

    def advance(self, *, timepoint=None, states=None, tissue=None):
        """
        A Ligand's next state is determined as follows:
        * its activator is ON
        * its ligand is currently ON in at least one neighboring tissue
        """
        assert states
        assert tissue

        activator_is_active = states.is_active(timepoint, self.delay,
                                               self.activator, tissue)

        ligands_in_neighbors = _retrieve_ligands(timepoint, states,
                                                 tissue, self.delay)

        ligand_present = False
        if self.ligand in ligands_in_neighbors:
            ligand_present = True

        activity = 0
        if activator_is_active and self.check_ligand(timepoint,
                                                     states,
                                                     tissue,
                                                     self.delay):
            activity = 1

        yield self.receptor, activity


class Gene:
    def __init__(self, *, name=None):
        global _genes

        assert name, "Gene must have a name"
        self.name = name

        _genes.append(self)
        self._set_ligand = None
        self._is_ligand = None

    def __repr__(self):
        return f"Gene('{self.name}')"

    def __eq__(self, other):
        return self.name == other.name

    def __ne__(self, other):
        return self.name != other.name

    def __lt__(self, other):
        return self.name < other.name

    def __hash__(self):
        return hash(self.name)

    def present(self):
        return 1

    def ligand_present(self):
        return (self._set_ligand is None) or 1 # @CTB

    def active(self):           # present = active
        if self._is_ligand:
            return self.present() and self.ligand_present()
        else:
            return self.present()

    def activated_by(self, *, source=None, delay=1):
        ix = Interaction_Activates(source=source, dest=self, delay=delay)
        _add_rule(ix)

    def activated_or(self, *, sources=None, delay=1):
        ix = Interaction_Or(sources=sources, dest=self, delay=delay)
        _add_rule(ix)

    def and_not(self, *, activator=None, repressor=None, delay=1):
        ix = Interaction_AndNot(source=activator, repressor=repressor,
                                dest=self, delay=delay)
        _add_rule(ix)

    def activated_by_and(self, *, sources, delay=1):
        ix = Interaction_And(sources=sources, dest=self, delay=delay)
        _add_rule(ix)

    def toggle_repressed(self, *, tf=None, cofactor=None, delay=1):
        ix = Interaction_ToggleRepressed(tf=tf, cofactor=cofactor,
                                         dest=self, delay=delay)
        _add_rule(ix)

    def is_present(self, *, where=None, start=None, duration=None):
        assert where
        assert start
        where.add_gene(gene=self, start=start, duration=duration)

    def custom_activation(self, *, state_fn=None, delay=1):
        ix = Interaction_Arbitrary(dest=self, state_fn=state_fn, delay=delay)
        _add_rule(ix)


class Ligand(Gene):
    def __init__(self, *, name=None):
        super().__init__(name=name)
        self._is_ligand = True

    def __repr__(self):
        return f"Ligand('{self.name}')"


class Receptor(Gene):
    def __init__(self, *, name=None, ligand=None):
        super().__init__(name=name)
        assert name
        self._set_ligand = ligand
        if ligand:
            ligand._is_ligand = True

    def __repr__(self):
        return f"Receptor('{self.name}')"

    def ligand(self, *, activator=None, ligand=None):
        # @CTB legacy
        if ligand is None:
            ligand = self._set_ligand
            if ligand is None:
                raise Exception("need to specify a ligand for this receptor, either at creation or here")
        else:
            ligand._is_ligand = True

        ix = Interaction_Ligand(activator=activator, ligand=ligand, receptor=self)
        _add_rule(ix)

    def activated_by(self, *, activator=None):
        assert self._set_ligand
        ix = Interaction_Ligand(activator=activator, ligand=ligand, receptor=self)
        _add_rule(ix)



import dinkum
from dinkum import vfg
from dinkum import vfn
from dinkum.vfg import Gene, Receptor
from dinkum.vfn import Tissue


def test_tissue_cmp():
    dinkum.reset()
    m = Tissue(name='M')
    n = Tissue(name='N')

    assert m < n


def test_gene_cmp():
    dinkum.reset()
    a = Gene(name='M')
    b = Gene(name='N')

    assert a < b


def test_gene_names_with_gene():
    dinkum.reset()
    a = Gene(name='M')
    b = Gene(name='N')
    assert vfg.get_gene_names() == ['M', 'N']


def test_gene_names_with_genes_and_receptor():
    dinkum.reset()
    c = Receptor(name='R')
    a = Gene(name='M')
    b = Gene(name='N')
    assert vfg.get_gene_names() == ['M', 'N', 'R']



import sys
from importlib.metadata import version

__version__ = version("dinkum-bio")

from . import vfg
from . import vfn
from . import observations

import itertools
import collections

class DinkumException(Exception):
    pass

class DinkumObservationFailed(DinkumException):
    pass


def reset(*, verbose=True):
    vfg.reset()
    vfn.reset()
    observations.reset()
    if verbose:
        print(f"initializing: dinkum v{__version__}")


def run_and_display(*, start=1, stop=10, gene_names=None, tissue_names=None,
            verbose=False):
    "@CTB document."
    from dinkum.display import MultiTissuePanel, tc_record_activity
    if not gene_names:
        gene_names = vfg.get_gene_names()
    if not tissue_names:
        tissue_names = vfn.get_tissue_names()

    try:
        states, tissues, is_active_fn = tc_record_activity(start=start,
                                                           stop=stop,
                                                           gene_names=gene_names,
                                                           verbose=verbose)
    except DinkumException as e:
        print(f"ERROR: {str(e)}", file=sys.stderr)
        print("Halting execution.", file=sys.stderr)
        return

    #@CTB what is tissues here and how might it differ from tissue_names?
    # perhaps just set it...
    #print(tissues)

    mp = MultiTissuePanel(states=states, tissue_names=tissue_names,
                          genes_by_name=gene_names)
    return mp.draw(is_active_fn)


class GeneActivity:
    def __init__(self):
        self.genes_by_name = {}

    def set_activity(self, *, gene=None, active=None):
        assert gene
        assert active is not None
        self.genes_by_name[gene.name] = active

    def is_active(self, gene_name):
        return self.genes_by_name.get(gene_name, False)

    def __contains__(self, gene):
        return self.is_active(gene.name)

    def report_activity(self):
        rl = []
        for k, v in sorted(self.genes_by_name.items()):
            if v:
                v = 1
            else:
                v = 0
            rl.append(f"{k}={v}")

        return rl


class State:
    """
    Hold the gene activity state for multiple tissues.

    Holds multiple tissue, each with their own GeneActivity object.

    dict interface supports getting and setting gene activity (value) by
    tissue (key).

    `tissues` attribute provides enumeration of available tissues

    `is_active(gene, tissue)` returns True/False around activity.
    """
    def __init__(self, *, tissues=None, time=None):
        assert tissues is not None
        assert time is not None

        self._tissues = list(tissues)
        self._tissues_by_name = {}
        self.time = time

    def __setitem__(self, tissue, genes):
        assert tissue in self._tissues
        assert isinstance(genes, GeneActivity)
        self._tissues_by_name[tissue.name] = genes

    def __getitem__(self, tissue):
        return self._tissues_by_name[tissue.name]

    def get_by_tissue_name(self, tissue_name):
        return self._tissues_by_name[tissue_name]

    @property
    def tissues(self):
        return self._tissues

    def is_active(self, gene, tissue):
        ts = self[tissue]
        if ts and gene in ts:
            return True
        return False


class States(collections.UserDict):
    """
    Contains (potentially incomplete) set of tissue/gene states for many
    timepoints.
    """
    def __init__(self):
        self.data = {}

    def is_active(self, current_tp, delay, gene, tissue):
        from .vfg import Gene

        assert int(current_tp)
        assert int(delay)
        assert isinstance(gene, Gene)

        check_tp = current_tp - delay
        state = self.get(check_tp)
        if state and state.is_active(gene, tissue):
            return True
        return False


class Timecourse:
    """
    Run a time course for a system b/t two time points, start and stop.
    """
    def __init__(self, *, start=None, stop=None):
        assert start is not None
        assert stop is not None

        print(f'start={start} stop={stop}')

        self.start = start
        self.stop = stop
        self.states_d = States()

    def __iter__(self):
        return iter(self.states_d.values())

    def __len__(self):
        return len(self.states_d)

    def run(self, *, verbose=False):
        start = self.start
        stop = self.stop

        tissues = vfn.get_tissues()
        if verbose:
            print(f"got {len(tissues)} tissues.")
            for t in tissues:
                print(f"\ttissue {t.name}")
            print('')

        # initialize state at start for all tissues
        this_state = State(tissues=tissues, time=start)
        for t in tissues:
            this_active = GeneActivity()
            for g in t.all_active(at=start):
                this_active.set_activity(gene=g, active=True)

            this_state[t] = this_active

        self.states_d[start] = this_state

        # advance one tick at a time
        for tp in range(start + 1, stop + 1):
            next_state = State(tissues=tissues, time=tp)

            for t in tissues:
                this_active = this_state[t]

                next_active = GeneActivity()

                # bring in maternal/always active
                seen = set()
                for g in t.all_active(at=tp):
                    next_active.set_activity(gene=g, active=1)
                    seen.add(g.name)

                for r in vfg.get_rules():
                    # advance state of all genes based on last state
                    for g, activity in r.advance(timepoint=tp,
                                                 states=self.states_d,
                                                 tissue=t):
                        if g.name in seen:
                            raise DinkumException(f"multiple rules containing {g.name}")
                        next_active.set_activity(gene=g, active=activity)
                        seen.add(g.name)


                next_state[t] = next_active

            # advance => next state
            self.states_d[tp] = next_state
            this_state = next_state

    def check(self):
        "Test all of the observations for all of the states."
        for state in iter(self):
            if not observations.test_observations(state):
                raise DinkumObservationFailed(state.time)


def run(start, stop, *, verbose=False):
    # run time course
    tc = Timecourse(start=start, stop=stop)
    tc.run(verbose=verbose)

    for state in tc:
        print(f"time={state.time}")
        for ti in state.tissues:
            present = state[ti]
            print(f"\ttissue={ti.name}, {present.report_activity()}")
        if not observations.test_observations(state):
            raise DinkumObservationFailed(state.time)

    return tc



from PIL import Image, ImageDraw


class PillowDrawer:
    def __init__(self, *, width=None, height=None, **kwargs):
        self.im = Image.new('RGB', (width, height), 'white')
        self.imdraw = ImageDraw.Draw(self.im)

    def polygon(self, coords, fill):
        self.imdraw.polygon(coords, fill=fill)

    def draw_text(self, text, xpos, ypos, color="black", align="center"):
        if align == "center":
            anchor = "lm"
        elif align == "right":
            anchor = "rm"
        else:
            raise Exception(f"unknown align string: '{align}'")
            
        self.imdraw.text((xpos, ypos), text, fill=color, anchor=anchor)
        
    def image(self):
        return self.im
